<?php

namespace NoccyLabs\Daemon;

use Psr\Log\LoggerTrait;

abstract class AbstractDaemon
{
    use LoggerTrait;

    /** @var bool True as long as the daemon is running, setting to false will cause shutdown */
    protected $isRunning = false;

    /** @var string Date format for the log */
    protected $dateFormat = 'M d H:i:s';
    
    /** @var array Loaded modules */
    protected $modules = [];
    
    /** @var array Module functions */
    protected $functions = [];

    /**
     * Set up the daemon
     *
     */
    abstract protected function onSetup();

    /**
     * Shutdown function
     *
     */
    abstract protected function onShutdown();

    /**
     * Loop function
     *
     */
    abstract protected function onLoop();

    /**
     * Signal handler
     *
     */
    public function onSignal($sigNo)
    {
        if ($sigNo == SIGINT) {
            $this->stop();
            echo "\n";
            $this->info("Received SIGINT");
        }
    }

    public function addModule(Module\AbstractModule $module)
    {
        try {
            $module->setDaemon($this);
            $module->onLoad();
        } catch (\Exception $e) {
            $this->error("Module failed to load: ".$module->getName()." - ".$e->getMessage());
        }
        $this->debug("Loaded module ".get_class($module));
        $this->modules[] = $module;
    }
    
    protected function updateModules()
    {
        foreach ($this->modules as $module) {
            $module->onTick();
        }
    }

    /**
     * {@inheritDoc}
     *
     */
    public function log($level, $message, array $context = array())
    {
        printf("%s [%s] %s\n", date($this->dateFormat), $level, $message);
    }

    public function registerFunction($name, callable $handler)
    {
        if (array_key_exists($name, $this->functions)) {
            throw new \Exception("Function {$name} is already registered");
        }
        $this->functions[$name] = $handler;
    }
    
    /**
     * Call a module function.
     * 
     * @param type $command
     * @param type $args
     * @return type
     * @throws \Exception
     */
    public function __call($name, $args)
    {
        if (!array_key_exists($name, $this->functions)) {
            throw new \Exception("No such callable registered: {$name}");
        }
        return call_user_func_array($this->functions[$name], $args);
    }
    
    /**
     * Start the daemon.
     *
     *
     */
    final public function run()
    {
        $this->onSetup();

        pcntl_signal(SIGINT, [ $this, "onSignal" ]);

        $this->isRunning = true;
        do {
            $this->onLoop();
            usleep(10000);
            $this->updateModules();
            pcntl_signal_dispatch();
        } while ($this->isRunning);

        $this->unloadModules();
        $this->onShutdown();
    }
    
    protected function unloadModules()
    {
        foreach ($this->modules as $module) {
            $module->onUnload();
        }
        $this->modules = [];
    }

    /**
     * Stop the daemon main loop.
     * 
     */
    final public function stop()
    {
        $this->isRunning = false;
    }

    /**
     * Static entry point.
     *
     */
    public static function main()
    {
        // Create a new instance of the requested daemon class
        $daemon = get_called_class();
        $inst = new $daemon();

        // Start the daemon
        $inst->run();
    }
}
